// @(#)OverView.h	2.21	10/29/91

#ifndef OverView_First

#ifdef __GNUG__
#pragma once
#endif

#define OverView_First

#include "View.h"
#include "Point.h"
#include "Shape.h"
#include "Dialog.h"

extern Point BADPOINT;


class AreaSelector: public MapCommand {
    MapView	*targetview;
public:
    MetaDef(AreaSelector);
    AreaSelector(MapView *mv, MapView *tv);
    void TrackFeedback(Point, Point, bool);
    void TrackConstrain(Point ap, Point pp, Point *np);
    Command *TrackMouse(TrackPhase, Point, Point, Point);
};


typedef void	(*pvf)();

class MapView: public View {
    float		MinLong, MaxLong;
    float		MinLat, MaxLat;
    float		LongSpan, LatSpan;
    float		MercLatSpan, MercMinLat, MercMaxLat;
    float		xfac, yfac;
    Point		extent;
    bool		new_content;
    bool		do_zoom_in, do_new_center;
    bool		in_load;
    Collection		*showcol;

    class MapDialogView *dialogView;
    GrCursor		oldcursor;
    class Dictionary	*rtreelayers;
    class Set		*failed_rtree;
    ImageCache		cache;

    void		DrawRtree(class TextItem *layer, Rectangle,
				bool extent_changed);
    void		DrawMemMap(TextItem *layer, Rectangle,
				bool extent_changed);
    void		DrawGrid(Rectangle);
    inline float	GridStep(float r);
    void		DrawLabel(Point pos, char *l);

protected:
    class Geo		*geo;
    class QuelQueries	*quel;

    bool		do_query_refresh;
    bool		do_display_refresh;

    // Overlay blinking:
    int			blink_count;
    int			blink_col;
    int			cur_blink_col;

    // Query Blinking:
    int			q_blink_col;
    TextItem		*q_blink_layer;

public:
    MetaDef(MapView);
    
    MapView(class Document*, Point, class ObjList*,
	float LongMin= 0, float LongMax= 0, float LatMin= 0, float LatMax= 0,
	byte *title= "Map Layerselect", bool overlay_flag= TRUE);
    ~MapView();
    
    //---- initializing
    void SetArea(float LongMin, float LongMax, float LatMin, float LatMax,
      bool init= FALSE);
    MapDialogView *GetDialogView()
	{ return dialogView; }
    Collection *GetMapOverlayCol();
    OrdCollection *GetQueryCol();
    virtual class Document *GetDocument() { return View::GetDocument(); }
    void SetGeo(Geo *g) { geo= g; }
    void NewQuel();

    class Window *GetBrowseWin();
    
    void InLoad() { in_load= TRUE; }
    void LoadingDone() { in_load= FALSE; }
    bool Loading() { return in_load; }

    //---- shape list management
    void Add(TextItem *);
    void LoadDone();

    void SetShowCol(class Collection *cp)
	{ showcol= cp; }

    //---- drawing
    virtual void TopDamaged() { AbstractMethod("TopDamaged"); }
    virtual void NewContent() { extent= gPoint0; }
    void Draw(Rectangle);
    virtual void BlinkQuery(TextItem *l);
    virtual void DrawBlinkQuery(Rectangle, TextItem *l, int col);
    Point GeoToPoint(float Long, float Lat);
    Point GeoToPoint(GeoPos&);
    void CompGeo();
    float XToLong(int);
    float YToLat(int);
    float LatToMerc(float);
    float CompLatToMerc(float);
    float AspectRatio() const { return float(extent.x) / extent.y; }
    Point CurExtent() { return extent; }
    void SetMapExtent(Point e) { extent= e; }
    void PToGeo(Point, GeoPos&);
    void DoZoomIn() { do_zoom_in= TRUE; }
    void ZoomInDone() { do_zoom_in= FALSE; }
    void DoZoomOut();
    void DoNewCenter() { do_new_center= TRUE; }
    void NewCenterDone() { do_new_center= FALSE; }
    bool ZoomingIn() { return do_zoom_in; }
    bool NewCenter() { return do_new_center; }
    bool IsGrabbing();
    bool AreaIsChanged() {return new_content; }
    virtual void DrawQueries(Rectangle);
    virtual void DisplayQuery(Rectangle r, class Collection *c);

    pvf			cs_ts;
    pvf			cs_fs;
    pvf			cs_constrain;
    pvf			cs_t;

    void		*user_ptr;
    
    void HourOn();
    void HourOff();

    float GetLongSpan()
	{ return LongSpan; }
    float GetLatSpan()
	{ return LatSpan; }
    float GetLongMin() { return MinLong; }
    float GetLongMax() { return MaxLong; }
    float GetLatMin() { return MinLat; }
    float GetLatMax() { return MaxLat; }
    char * GetQBbox();
    Geo *GetGeo()
	{ return geo; }
    QuelQueries *GetQueries()
	{ return quel; }
    class istream &LoadQueries(istream &);

    virtual void DoQueryRefresh();
    void QDisplayRefresh() { do_display_refresh= TRUE; }

    void SetCoordSys(char *s, pvf ts= 0, pvf fs= 0, pvf constrain= 0, pvf t= 0);
    void ShowPos(Point p);

    //---- event handling
    Command *DoLeftButtonDownCommand(Point, Token, int) { return gNoChanges; }
    Command *DoMiddleButtonDownCommand(Point, Token, int);
    Command *DoOtherEventCommand(Point, Token t);
    Command *DoIdleCommand();

    ostream& Save(ostream&);    
    istream& Load(istream&);
};


class OverView: public MapView {
float       area_long_min,
	    area_long_max, 
	    area_lat_min, 
	    area_lat_max;
MapView	    *mapview;
protected:
    Rectangle OverView::GeoToRect(float area_long_min, float area_long_max,
    	float area_lat_min, float area_lat_max);
public:
    MetaDef(OverView);
    OverView(class Document*, Point, class ObjList*, MapView*,
	float LongMin= 0, float LongMax= 0, float LatMin= 0, float LatMax= 0,
	byte *title= "Overview Layerselect");

    float AreaLongMin() { return area_long_min; }
    float AreaLongMax() { return area_long_max; }
    float AreaLatMin() { return area_lat_min; }
    float AreaLatMax() { return area_lat_max; }
    
    class Document *GetDocument() { return mapview->GetDocument(); }
    void Draw(Rectangle);
    void ShowArea(Rectangle, float, float, float, float);

    Command *DoMiddleButtonDownCommand(Point, Token, int);
    ostream& Save(ostream&);    
    istream& Load(istream&);
};


class MapDialogView: public DialogView {
    class MapLayerDialog *layerdialog,
			 *overlaydialog,
			 *querydialog;
    class TextItem	*Long, *Lat, *Coordsys;
    class ToggleButton	*show_grid;
    class ToggleButton	*show_labels;
    class Layer		*top_overlay;
    class Layer		*top_query;
    MapView		*mapView;
    bool		order_changed;
    int			overlays;
    class ZoomDialog	*zd;
public:
    MetaDef(MapDialogView);

    MapDialogView(MapView *, byte *title, bool overlay_flag);
    ~MapDialogView();
    void Control(int id, int p, void *v);
    VObject *DoCreateDialog();

    MapLayerDialog *GetLayerDialog() { return layerdialog; }
    MapLayerDialog *GetOverlayDialog()
	{ return overlays ? overlaydialog: 0; }
    MapLayerDialog *GetQueryDialog()
	{ return querydialog; }
    void LayersChanged(class LayerDialog *l);
    bool OrderChanged() { return order_changed; }
    bool ShowGrid();
    bool ShowLabels();
    MapView *GetMapView() const { return mapView; }
    void SetTopOverlay();
    void SetTopQuery();

    SetCoordSys(char *s) { Coordsys->SetString(s, TRUE); UpdateEvent(); }
    SetLong(char *s) {Long->SetString(s, TRUE); UpdateEvent(); }
    SetLat(char *s) {Lat->SetString(s, TRUE); UpdateEvent(); }

    ostream& Save(ostream&);    
    istream& Load(istream&);
};

#endif OverView_First
